#!/usr/bin/env python3

"""
ticket helper utility
credits: godreborn 

"""
import os
import json

class WiiUTicketHelper:
    tik = {
        "start": b'\x00\x01\x00\x04',
        "fake" : b'\xD1\x5E\xA5\xED\x15\xAB\xE1\x1A',
        "size" : 0x100,
        "pos_tid" : 476,
        "tik_lens" : [ 696, 848 ]
    }
    tik["len_start_bytes"] = len(tik["start"]) 
    VERBOSE=False
    def __init__(self,fn,include_ticket_data=False,json_title_file=None,add_title_key=False,
                 replace_title_name=True,replace_str=("\n"," ")):
        """
        with include_ticket_data = True:
        ticket (dictonary): index (int), [ticket_data (bytes), title_id (str), fake/valid ticket (str), start, end, length (all int)] (list)
    
        without include_ticket_data:
        ticket (dictonary): index (int), [title_id (str), fake/valid ticket (str), start, end, length (all int)] (list)
        """
        self.fn = os.path.abspath(fn)
        self.f = open(fn,"r+b")
        self.pos = 0
        self.filesize = os.stat(self.fn).st_size
        for length in self.tik["tik_lens"]:
            if self.filesize % length == 0:
                self.same_size_tickets = length
                break
            else:
                self.same_size_tickets = None
        if self.VERBOSE: self.v_print("same size tickets",self.same_size_tickets)
        self.index = 0
        self.tickets = {}
        self.include_ticket_data = include_ticket_data
        self.add_title_key = add_title_key
        self.json_title_file = self.read_json(json_title_file) if json_title_file is not None else json_title_file
        self.replace_title_name = replace_title_name
        self.replace_str = replace_str
        
    def v_print(self,msg,val,txtfmt=None):
        FMT = { "red":"\x1b[31m","green":"\x1b[32m","bold":"\x1b[1m","cyan":"\x1b[36m","default":"\x1b[0m" }
        if txtfmt is None:
            print("\n%s %s (type: %s)\n" %(msg,val,type(val)))
        else:
            print("%s\n%s %s (type: %s)\n%s" %(FMT[txtfmt],msg,val,type(val),FMT["default"]))
            
    def __enter__(self):
        return(self)
    
    def is_valid_ticket(self,search_bytes,keep_pos=False):
        if isinstance(search_bytes,bytes):
            if keep_pos is False:
                self.pos = self.f.tell()
            if self.VERBOSE: self.v_print("FILE POSTION",self.pos,txtfmt="bold");
            if self.tik["start"] == search_bytes:
                if self.VERBOSE: self.v_print("ticket found!, bytes left:",self.filesize-self.pos,txtfmt="green");
                return(True)
            else:
                return(False)
        else:
            raise TypeError("%s needed not %s" %(bytes,type(search_bytes)))
        
    def read_json(self,fn):
        with open(os.path.abspath(fn)) as f:
            return(json.load(f))
        
    def search_ticket_len(self):
        if self.same_size_tickets is not None:
            # check valid ticket start
            peek_tik = self.f.peek()[0:self.tik["len_start_bytes"]]
            return(self.same_size_tickets if self.is_valid_ticket(peek_tik,keep_pos=True) else None)
        
        if self.filesize - self.pos in self.tik["tik_lens"]:
            # only last ticket left
            last_ticket_len = self.filesize - self.pos
            if self.VERBOSE: self.v_print("EXPECTED LAST TICKET LEN",last_ticket_len,"bold");
            peek_tik = self.f.peek(self.filesize)[0:self.tik["len_start_bytes"]]
            return(last_ticket_len if self.is_valid_ticket(peek_tik,keep_pos=True) else None)

        for length in self.tik["tik_lens"]:
            # check in case of broken ticket with e.g. wrong filesize (last ticket size not in tik_lens)
            if self.pos + length > self.filesize:
                raise ValueError("current positon (%i) + ticket length (%i) > filesize (%i)" %(self.pos,length,self.filesize))
            start, end = length, length+self.tik["len_start_bytes"]
            peek_tik = self.f.peek()[start:end]
            if self.VERBOSE: self.v_print("START BYTES NEXT TICKET CHECK",peek_tik,txtfmt="red")
            if self.is_valid_ticket(peek_tik,keep_pos=True):
                if self.VERBOSE: self.v_print("Ticket found with length",length,txtfmt="green");
                tik_len = length
                break
            else:
                if self.VERBOSE: self.v_print("Ticket not found with length",length,txtfmt="red");
                tik_len = None
        return(tik_len)

    def get_next_ticket(self):
        if self.pos >= self.filesize:
            self.v_print("END OF FILE REACHED",self.pos)
            return(False)
        #ensure correct position
        self.f.seek(self.pos)
        tik_len = self.search_ticket_len()
        if isinstance(tik_len,int) is not True:
            # should never occur
            self.v_print("",tik_len,txtfmt="red")
            raise TypeError( "expected int, got %s" %(type(tik_len)) )
        tmp_tik = self.f.read(tik_len)
        if self.VERBOSE: self.v_print("",tmp_tik);
        self.tickets[self.index] = [self.get_ticket_id(tmp_tik),self.check_fake_ticket(tmp_tik),self.pos,self.pos+tik_len,tik_len]
        if self.json_title_file is not None:
            tmp_title_dic = self.get_title_name_key(self.tickets[self.index][0]) # tik id
            for i in tmp_title_dic:
                self.tickets[self.index].insert(1,i) # insert after ticket id
        if self.include_ticket_data:
            self.tickets[self.index].insert(0,tmp_tik)
        self.pos = self.f.tell()
        self.index += 1
        
    def return_tickets(self):
        return(self.tickets)
        
    def get_all_tickets(self):
        while self.pos < self.filesize:
            self.get_next_ticket()
            
    def get_ticket_id(self,ticket_data,join_tid=True,joiner=""):
        if isinstance(ticket_data,bytes) is not True:
            raise TypeError( "expected bytes, got %s" %(type(ticket_data)) )
        else:
            tid = [ "{:02x}".format(i) for i in ticket_data[self.tik["pos_tid"]:self.tik["pos_tid"]+8] ]
            return(joiner.join(tid) if join_tid else tid)
        
    def check_fake_ticket(self,ticket_data):
        x = int(self.tik["size"]/len(self.tik["fake"]))
        if x * self.tik["fake"] in ticket_data[self.tik["len_start_bytes"]:self.tik["len_start_bytes"]+self.tik["size"]]:
            return("FAKE")
        else:
            return("VALID")
    
    def get_title_name_key(self,tik_id):
        res = [ None ]
        if self.add_title_key: res.append(None);
        for i in range(0,len(self.json_title_file)):
            if self.json_title_file[i]["titleID"] in tik_id:
                res = [ self.json_title_file[i]['name'] ]
                if self.replace_title_name:
                    try:
                        res[0] = res[0].replace(self.replace_str[0],self.replace_str[1])
                    except AttributeError as e:
                        if res[0] is None:
                            pass
                        else:
                            self.v_print("",str(e),"red")
                if self.add_title_key: res.append(self.json_title_file[i]['titleKey']);
                break
        if self.VERBOSE: self.v_print("TITLE NAME",res[0],"green");
        if self.VERBOSE and self.add_title_key: self.v_print("TITLE KEY",res[1],"green");
        res.reverse()
        return(res)
        
    def return_ticket_gen(self):
        gen_index = 0
        while gen_index < self.index:
            yield(self.tickets[gen_index])
            gen_index += 1
        
    def __exit__(self,val,typ,traceback):
        self.f.close()

